/* nc100em, an Amstrad NC100/NC200 emulator.
 * Copyright (C) 1995-2022 Russell Marks and Ian Collier.
 * See common.c for licence details.
 *
 * sdlmain.c - SDL2 version.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>

#include <SDL2/SDL.h>

#include "common.h"
#include "z80.h"

#include "sdlicon.h"


/* width of the border (on all sides) around the actual screen image.
 * (Must be <=15 given how update_scrn() works.)
 */
#define BORDER_WIDTH		5


char *cmd_name="dnc100em";

int scale;
int hsize,vsize;

SDL_Window *window=NULL;
SDL_Renderer *renderer=NULL;
SDL_Texture *texture=NULL;
int is332=1;

int black_r,black_g,black_b;
int white_r,white_g,white_b;


void exit_program(void)
{
if(texture) SDL_DestroyTexture(texture);
if(renderer) SDL_DestroyRenderer(renderer);
if(window) SDL_DestroyWindow(window);
SDL_Quit();

common_uninit();

if(write_ram_on_close)
  writeram(mem+RAM_START);
writecard(cardmem);
exit(0);
}


/* XXX not doing the co-op mouse support initially */
void get_mouse_state(int *mxp,int *myp,int *mbutp)
{
*mxp=*myp=*mbutp=0;
}


void key_pressrel(int port,int mask,int press)
{
if(press)
  keyports[port]|=mask;
else
  keyports[port]&=~mask;
}


void key_event(SDL_Scancode sc,int press)
{
switch(sc)
  {
  case SDL_SCANCODE_F5:
  case SDL_SCANCODE_F10:
    if(press) do_nmi=1;
    break;
  case SDL_SCANCODE_F8:
    if(press) exit_program();	/* doesn't return */
    break;
  
  /* Insert = Function, but allow F1/"GUI" to act as Function too. */
  case SDL_SCANCODE_INSERT:
  case SDL_SCANCODE_F1:
  case SDL_SCANCODE_LGUI: case SDL_SCANCODE_RGUI:
    key_pressrel(1,0x01,press); break;

  case SDL_SCANCODE_RETURN: 
    key_pressrel(0,0x10,press); break;
  case SDL_SCANCODE_LCTRL:
    key_pressrel(1,0x02,press); break;
  case SDL_SCANCODE_LSHIFT:
    key_pressrel(0,0x01,press); break;
  case SDL_SCANCODE_RSHIFT:
    key_pressrel(0,0x02,press); break;
  case SDL_SCANCODE_LALT: case SDL_SCANCODE_RALT:
    key_pressrel(2,0x02,press); break;
  case SDL_SCANCODE_CAPSLOCK:
    key_pressrel(2,0x01,press); break;
  case SDL_SCANCODE_BACKSPACE: case SDL_SCANCODE_DELETE:
    key_pressrel(9,0x04,press); break;
  case SDL_SCANCODE_ESCAPE:
    key_pressrel(1,0x04,press); break;
  case SDL_SCANCODE_TAB:
    key_pressrel(2,0x08,press); break;
  case SDL_SCANCODE_UP:
    key_pressrel(7,0x08,press); break;
  case SDL_SCANCODE_DOWN:
    key_pressrel(6,0x02,press); break;
  case SDL_SCANCODE_LEFT:
    key_pressrel(0,0x08,press); break;
  case SDL_SCANCODE_RIGHT:
    key_pressrel(6,0x08,press); break;
  case SDL_SCANCODE_MINUS:
    key_pressrel(8,0x02,press); break;
  case SDL_SCANCODE_EQUALS:
    if(nc200) key_pressrel(7,0x02,press); else key_pressrel(7,0x01,press);
    break;
  case SDL_SCANCODE_SEMICOLON:
    key_pressrel(9,0x10,press); break;
  case SDL_SCANCODE_APOSTROPHE:
    key_pressrel(8,0x10,press); break;
  case SDL_SCANCODE_COMMA:
    key_pressrel(8,0x80,press); break;
  case SDL_SCANCODE_PERIOD:
    key_pressrel(9,0x80,press); break;
  case SDL_SCANCODE_SLASH:
    key_pressrel(6,0x20,press); break;
  case SDL_SCANCODE_APPLICATION:	/* menu key for menu? surely not */
  case SDL_SCANCODE_GRAVE:
    key_pressrel(7,0x10,press); break;

  case SDL_SCANCODE_END:
    /* an odd mapping, but keeps "Del->" available */
    key_pressrel(6,0x04,press); break;

  case SDL_SCANCODE_BACKSLASH:
    /* SDL2 seems to use BACKSLASH to refer to a key next to CR, as
     * usual for a US layout. We actually want that key to do # on
     * the NC.
     */
    key_pressrel(6,0x10,press); break;
  case SDL_SCANCODE_NONUSBACKSLASH:
  case SDL_SCANCODE_HOME:		/* for e.g. US/JP layouts */
    key_pressrel(7,0x04,press); break;

    /* not using keytable here, so some more in ASCII order */
  case SDL_SCANCODE_SPACE:
    key_pressrel(1,0x08,press); break;
  
  case SDL_SCANCODE_0:
    if(nc200) key_pressrel(9,0x02,press); else key_pressrel(9,0x01,press);
    break;
  case SDL_SCANCODE_1:
    key_pressrel(2,0x04,press); break;
  case SDL_SCANCODE_2:
    key_pressrel(3,0x02,press); break;
  case SDL_SCANCODE_3:
    key_pressrel(3,0x01,press); break;
  case SDL_SCANCODE_4:
    if(nc200) key_pressrel(0,0x04,press); else key_pressrel(4,0x01,press);
    break;
  case SDL_SCANCODE_5:
    if(nc200) key_pressrel(2,0x10,press); else key_pressrel(1,0x40,press);
    break;
  case SDL_SCANCODE_6:
    if(nc200) key_pressrel(2,0x40,press); else key_pressrel(6,0x01,press);
    break;
  case SDL_SCANCODE_7:
    if(nc200) key_pressrel(4,0x02,press); else key_pressrel(7,0x02,press);
    break;
  case SDL_SCANCODE_8:
    if(nc200) key_pressrel(4,0x01,press); else key_pressrel(8,0x01,press);
    break;
  case SDL_SCANCODE_9:
    if(nc200) key_pressrel(1,0x80,press); else key_pressrel(9,0x02,press);
    break;

  case SDL_SCANCODE_A:
    key_pressrel(4,0x10,press); break;
  case SDL_SCANCODE_B:
    key_pressrel(5,0x04,press); break;
  case SDL_SCANCODE_C:
    key_pressrel(5,0x80,press); break;
  case SDL_SCANCODE_D:
    key_pressrel(3,0x80,press); break;
  case SDL_SCANCODE_E:
    key_pressrel(3,0x10,press); break;
  case SDL_SCANCODE_F:
    key_pressrel(4,0x80,press); break;
  case SDL_SCANCODE_G:
    key_pressrel(5,0x40,press); break;
    
  case SDL_SCANCODE_H:
    key_pressrel(6,0x40,press); break;
  case SDL_SCANCODE_I:
    key_pressrel(8,0x20,press); break;
  case SDL_SCANCODE_J:
    key_pressrel(8,0x40,press); break;
  case SDL_SCANCODE_K:
    key_pressrel(7,0x80,press); break;
  case SDL_SCANCODE_L:
    key_pressrel(9,0x20,press); break;
  case SDL_SCANCODE_M:
    key_pressrel(7,0x40,press); break;
  case SDL_SCANCODE_N:
    key_pressrel(6,0x80,press); break;
  case SDL_SCANCODE_O:
    key_pressrel(9,0x40,press); break;
    
  case SDL_SCANCODE_P:
    key_pressrel(9,0x08,press); break;
  case SDL_SCANCODE_Q:
    key_pressrel(3,0x04,press); break;
  case SDL_SCANCODE_R:
    key_pressrel(4,0x40,press); break;
  case SDL_SCANCODE_S:
    key_pressrel(3,0x40,press); break;
  case SDL_SCANCODE_T:
    key_pressrel(5,0x10,press); break;
  case SDL_SCANCODE_U:
    key_pressrel(7,0x20,press); break;
  case SDL_SCANCODE_V:
    key_pressrel(5,0x08,press); break;
  case SDL_SCANCODE_W:
    key_pressrel(3,0x08,press); break;
    
  case SDL_SCANCODE_X:
    key_pressrel(4,0x08,press); break;
  case SDL_SCANCODE_Y:
    key_pressrel(5,0x20,press); break;
  case SDL_SCANCODE_Z:
    key_pressrel(4,0x04,press); break;
  case SDL_SCANCODE_LEFTBRACKET:
    key_pressrel(8,0x08,press); break;
  case SDL_SCANCODE_RIGHTBRACKET:
    key_pressrel(8,0x04,press); break;

  default:
    /* yes gcc, I know I am not handling every key in existence */
    ;
  }
}


/* redraw the screen. With the way SDL2 works, differential updating
 * seems pointless to even attempt as SDL pretty much forces a
 * full-redraw approach.
 */
void update_scrn(void)
{
static unsigned char lookup[256*32];
static int count=0,first=1;
unsigned char *ramptr;
unsigned char *dst;
unsigned char *imagedata;
int force=0;
int linelen;
int x,y,mr,mg,mb,mask;
int lines=nc200?128:64;

if(first)
  {
  force=1,first=0;

  /* generate image data lookup table for input bytes */
  for(x=0;x<256;x++)
    {
    dst=lookup+32*x;
    for(mask=128;mask;mask>>=1)
      {
      if(x&mask)
        mr=black_r,mg=black_g,mb=black_b;
      else
        mr=white_r,mg=white_g,mb=white_b;
      
      if(is332)
        *dst++=mr;	/* just uses "red" */
      else
        {
        /* supposedly this is how you're meant to do this */
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
        *dst++=255;
        *dst++=mb;
        *dst++=mg;
        *dst++=mr;
#else
        *dst++=mr;
        *dst++=mg;
        *dst++=mb;
        *dst++=255;
#endif
        }
      }
    }
  }

/* only do it every 1/Nth
 * (non-zero `force' overrides this)
 */
count++;
if(count<scrn_freq && !force) return;

count=0;

if(SDL_LockTexture(texture,NULL,(void **)&imagedata,&linelen)!=0)
  fprintf(stderr,"%s: texture lock failed.\n",cmd_name),exit_program();

/* draw top line of border */
if(is332)	/* 1-byte pixels */
  {
  memset(imagedata,*lookup,linelen);
  }
else		/* 4-byte pixels */
  {
  /* XXX this isn't ideal, even if the memcpy() gets optimised */
  dst=imagedata;
  for(x=0;x<linelen;x+=4,dst+=4)
    memcpy(dst,lookup,4);
  }

/* copy the top line to other lines in the border */
for(y=0;y<BORDER_WIDTH;y++)
  {
  if(y) memcpy(imagedata+y*linelen,imagedata,linelen);
  memcpy(imagedata+(BORDER_WIDTH+lines+y)*linelen,imagedata,linelen);
  }

ramptr=mem+RAM_START+scrnaddr;

/* draw NC screen (plus left/right border), using lookup array for
 * each byte.
 */
for(y=0;y<lines;y++,ramptr+=4)
  {
  dst=imagedata+linelen*(y+BORDER_WIDTH);
  
  /* using constant memcpy() sizes gives the compiler a chance of
   * inlining the copy, rather than actually calling memcpy().
   * (Tested with gcc - it's 100% memcpy() calls without this
   * approach.)
   */
  if(is332)	/* 1-byte pixels */
    {
    memcpy(dst,lookup,BORDER_WIDTH);
    dst+=BORDER_WIDTH;
    for(x=0;x<480;x+=8,dst+=8)
      memcpy(dst,lookup+32*(*ramptr++),8);
    memcpy(dst,lookup,BORDER_WIDTH);
    }
  else		/* 4-byte pixels */
    {
    memcpy(dst,lookup,BORDER_WIDTH*4);
    dst+=BORDER_WIDTH*4;
    for(x=0;x<480;x+=8,dst+=32)
      memcpy(dst,lookup+32*(*ramptr++),32);
    memcpy(dst,lookup,BORDER_WIDTH*4);
    }
  }

SDL_UnlockTexture(texture);

/* not really used here, but reset it anyway */
force_full_redraw=0;
}


void do_interrupt(void)
{
static int scount=0;
SDL_Event event;

update_scrn();

SDL_RenderClear(renderer);
SDL_RenderCopy(renderer,texture,NULL,NULL);
SDL_RenderPresent(renderer);

while(SDL_PollEvent(&event))
  {
  switch(event.type)
    {
    case SDL_QUIT:
      do_nmi=1;
      break;
    case SDL_KEYDOWN:
      key_event(event.key.keysym.scancode,1);
      break;
    case SDL_KEYUP:
      key_event(event.key.keysym.scancode,0);
      break;
    }
  }

/* XXX not sure how sensible this is */
scount++;
if(scount==4) scount=0,serout_flush();
}


void initwindow(void)
{
SDL_Surface *surface;
int lines,w,h,tmp;
int try332=1;

/* XXX try332 is never currently set to anything else, but that might
 * change in future (e.g. if a manual override option is added).
 */

/* set scale from default in common.c */
lines=(nc200?128:64);
scale=default_scale;
w=2*BORDER_WIDTH+480; h=2*BORDER_WIDTH+lines;
hsize=w*scale; vsize=h*scale;

black_r=black_g=black_b=0;
white_r=white_g=white_b=255;

if(invert_video)
  {
  /* swap black/white */
  tmp=black_r; black_r=white_r; white_r=tmp;
  tmp=black_g; black_g=white_g; white_g=tmp;
  tmp=black_b; black_b=white_b; white_b=tmp;
  }

SDL_Init(SDL_INIT_VIDEO);

if((window=SDL_CreateWindow(nc200?"dnc200em":(nc150?"dnc150em":"dnc100em"),
                            SDL_WINDOWPOS_UNDEFINED,SDL_WINDOWPOS_UNDEFINED,
                            hsize,vsize,0))==NULL)
  fprintf(stderr,"%s: couldn't create window.\n",cmd_name),
    exit_program();

/* set icon if possible */
if((surface=SDL_CreateRGBSurfaceWithFormatFrom((void *)dnc100em_icon_data,
                                               DNC100EM_ICON_WIDTH,
                                               DNC100EM_ICON_HEIGHT,
                                               24,3*DNC100EM_ICON_WIDTH,
                                               SDL_PIXELFORMAT_RGB24))!=NULL)
  {
  SDL_SetWindowIcon(window,surface);
  SDL_FreeSurface(surface);
  }

/* first try accelerated, then anything else */
if((renderer=SDL_CreateRenderer(window,-1,
                                SDL_RENDERER_ACCELERATED))==NULL)
  {
  if((renderer=SDL_CreateRenderer(window,-1,0))==NULL)
    fprintf(stderr,"%s: couldn't create renderer.\n",cmd_name),
      exit_program();
  }

/* any scaling is done by SDL, assuming it will be faster */
SDL_SetHint(SDL_HINT_RENDER_SCALE_QUALITY,"nearest");
SDL_RenderSetLogicalSize(renderer,w,h);

/* the assumption here is that RGB332 will be faster due to being
 * much smaller, but that might be questionable - it should make the
 * software side faster at least. It's somewhat faster when tested
 * at higher scales (like 6) on a Raspberry Pi 400.
 */
if(!try332 ||
   (texture=SDL_CreateTexture(renderer,SDL_PIXELFORMAT_RGB332,
                              SDL_TEXTUREACCESS_STREAMING,w,h))==NULL)
  {
  is332=0;
  if((texture=SDL_CreateTexture(renderer,SDL_PIXELFORMAT_RGBA32,
                                SDL_TEXTUREACCESS_STREAMING,w,h))==NULL)
    fprintf(stderr,"%s: texture creation failed.\n",cmd_name),
      exit_program();
  }

if(is332)
  {
  /* XXX not sure if any bit-ordering issues possible with RGB332.
   * But I suppose for pure black/white it won't matter. :-)
   */
  black_r=((black_r&0xe0)|((black_g&0xe0)>>3)|(black_b>>6));
  white_r=((white_r&0xe0)|((white_g&0xe0)>>3)|(white_b>>6));
  }
}


int main(int argc,char *argv[])
{
parseoptions(argc,argv);

common_init();

initwindow();

startsigsandtimer();
mainloop();

/* not reached */
exit(0);
}
