/* nc100em, an Amstrad NC100/NC200 emulator.
 * Copyright (C) 1995-2001 Russell Marks and Ian Collier.
 * See common.c for licence details.
 *
 * sound.c - the sound emulation, based on the sound chip emulation I
 *           wrote for Fuse. The OSS code is partly due to Matan Ziv-Av,
 *           and Philip Kendall probably frobbed some stuff too. :-)
 */

/* NB: since it's less than clear what the NC100's sound chip (or rather,
 * whatever it is that actually generates the sound) is called, I've
 * left various bits here with "ay_" in the name. They don't have anything
 * to do with an AY sound chip here (I'd have thought), but since I don't
 * have a better name... :-)
 */

#ifdef OSS_SOUND_SUPPORT

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <sys/soundcard.h>
#include "common.h"

#include "sound.h"

/* configuration */
int sound_enabled=0;
int sound_freq=44100;
int sound_stereo=0;		/* true for stereo *output sample* (only) */

/* It's true that we could get away with a much lower sample rate for
 * most things (11025, perhaps), but it needs to be pretty high to
 * reasonably emulate e.g. play4b sample-playing. And it's only mono
 * 8-bit, and sound is disabled by default - so it's not *that* big a
 * burden IMHO.
 */


/* Must be <=127 for all channels; 2*24 = 48.
 * (Since they're harsh square waves used mainly for beeping, we
 * keep the level much lower than 127 :-))
 */
#define AMPL_AY_TONE		24	/* two of these */

/* max. number of sub-frame port writes allowed;
 * given the number of port writes theoretically possible in a
 * 100th I think this should be plenty.
 */
#define AY_CHANGE_MAX		4000

static int sound_framesiz;

static unsigned char *sound_buf;

/* ay_tone_subcycles is fixed-point with low 16 bits as fractional part.
 */
static unsigned int ay_tone_tick[2],ay_tone_high[2];
static unsigned int ay_tone_subcycles;
static unsigned int ay_tick_incr;
static unsigned int ay_tone_period[2];

/* sound chip registers (?) - last writes to ports 50h-53h. */
static unsigned char sound_ay_registers[4];

struct ay_change_tag
  {
  unsigned long tstates;
  unsigned short ofs;
  unsigned char reg,val;
  };

static struct ay_change_tag ay_change[AY_CHANGE_MAX];
static int ay_change_count;


static int soundfd=-1;
static int sixteenbit=0;



/* using (8) 64 byte frags for 8kHz, scale up for higher */
#define BASE_SOUND_FRAG_PWR	6



/* returns non-zero on success, and adjusts freq/stereo args to reflect
 * what we've actually got.
 */
int osssound_init(int *freqptr,int *stereoptr)
{
int frag,tmp;

if((soundfd=open("/dev/dsp",O_WRONLY))<0)
  return(0);

tmp=AFMT_U8;
if(ioctl(soundfd,SNDCTL_DSP_SETFMT,&tmp)==-1)
  {
  /* try 16-bit - may be a 16-bit-only device */
  tmp=AFMT_S16_LE;
  if((ioctl(soundfd,SNDCTL_DSP_SETFMT,&tmp))==-1)
    {
    close(soundfd);
    return(0);
    }

  sixteenbit=1;
  }

tmp=(*stereoptr)?1:0;
if(ioctl(soundfd,SNDCTL_DSP_STEREO,&tmp)<0)
  {
  /* if it failed, make sure the opposite is ok */
  tmp=(*stereoptr)?0:1;
  if(ioctl(soundfd,SNDCTL_DSP_STEREO,&tmp)<0)
    {
    close(soundfd);
    return(0);
    }
  *stereoptr=tmp;
  }

frag=(0x80000|BASE_SOUND_FRAG_PWR);

if(ioctl(soundfd,SNDCTL_DSP_SPEED,freqptr)<0)
  {
  close(soundfd);
  return(0);
  }

if(*freqptr>8250) frag++;
if(*freqptr>16500) frag++;
if(*freqptr>33000) frag++;
if(*stereoptr) frag++;
if(sixteenbit) frag++;

if(ioctl(soundfd,SNDCTL_DSP_SETFRAGMENT,&frag)<0)
  {
  close(soundfd);
  return(0);
  }

return(1);	/* success */
}


void osssound_end(void)
{
if(soundfd!=-1)
  close(soundfd);
}


void osssound_frame(unsigned char *data,int len)
{
static unsigned char buf16[8192];
int ret=0,ofs=0;

if(sixteenbit)
  {
  unsigned char *src,*dst;
  int f;

  src=data; dst=buf16;
  for(f=0;f<len;f++)
    {
    *dst++=128;
    *dst++=*src++-128;
    }

  data=buf16;
  len<<=1;
  }

while(len)
  {
  ret=write(soundfd,data+ofs,len);
  if(ret>0)
    ofs+=ret,len-=ret;
  }
}


void sound_ay_init(void)
{
int f;

sound_ay_registers[0]=sound_ay_registers[2]=0;
sound_ay_registers[1]=sound_ay_registers[3]=0x80;

ay_tone_subcycles=0;
for(f=0;f<2;f++)
  ay_tone_tick[f]=ay_tone_high[f]=0,ay_tone_period[f]=1;

/* nciospec.doc says frequency=1000000/(data*2*1.6276))
 * which arguably suggests that the clock is 307200Hz.
 * It seems plausible in BBC Basic at least. But it's hard
 * to tell past the massive distortion on the real NC.
 * I was close to going with 614400Hz instead, as the
 * distortion is so bad it's hard to even tell what *octave*
 * a tone is. :-)
 */
ay_tick_incr=(int)(65536.*307200./sound_freq);

ay_change_count=0;
}


int sound_init(void)
{
if(!osssound_init(&sound_freq,&sound_stereo))
  return(0);

sound_enabled=1;
sound_framesiz=sound_freq/100;

if((sound_buf=malloc(sound_framesiz*(sound_stereo+1)))==NULL)
  {
  sound_end();
  return(0);
  }

sound_ay_init();

return(1);
}


void sound_end(void)
{
if(sound_enabled)
  {
  if(sound_buf)
    free(sound_buf);
  osssound_end();
  sound_enabled=0;
  }
}



#define AY_GET_SUBVAL(chan)	(level*2*ay_tone_tick[chan]/tone_count)

/* XXX not great having this as a macro to inline it */
#define AY_DO_TONE(var,chan) \
  (var)=0;								\
  is_low=0;								\
  if(level)								\
    {									\
    if(ay_tone_high[chan])						\
      (var)= (level);							\
    else								\
      (var)=-(level),is_low=1;						\
    }									\
  									\
  ay_tone_tick[chan]+=tone_count;					\
  count=0;								\
  while(ay_tone_tick[chan]>=ay_tone_period[chan])			\
    {									\
    count++;								\
    ay_tone_tick[chan]-=ay_tone_period[chan];				\
    ay_tone_high[chan]=!ay_tone_high[chan];				\
    									\
    /* has to be here, unfortunately... */				\
    if(count==1 && level && ay_tone_tick[chan]<tone_count)		\
      {									\
      if(is_low)							\
        (var)+=AY_GET_SUBVAL(chan);					\
      else								\
        (var)-=AY_GET_SUBVAL(chan);					\
      }									\
    }									\
  									\
  /* if it's changed more than once during the sample, we can't */	\
  /* represent it faithfully. So, just hope it's a sample.      */	\
  /* (That said, this should also help avoid aliasing noise.)   */	\
  if(count>1)								\
    (var)=-(level)


static void sound_ay_overlay(void)
{
int f,level,count;
unsigned char *ptr;
struct ay_change_tag *change_ptr=ay_change;
int changes_left=ay_change_count;
int reg,r;
int is_low;
int chan1,chan2;
int tsm100=tsmax*100;
unsigned int tone_count;

/* convert change times to sample offsets */
for(f=0;f<ay_change_count;f++)
  ay_change[f].ofs=(ay_change[f].tstates*sound_freq)/tsm100;

for(f=0,ptr=sound_buf;f<sound_framesiz;f++)
  {
  while(changes_left && f>=change_ptr->ofs)
    {
    sound_ay_registers[reg=change_ptr->reg]=change_ptr->val;
    change_ptr++; changes_left--;

    r=reg>>1;
    /* a zero-len period is the same as 1 */
    ay_tone_period[r]=(sound_ay_registers[reg&~1]|
                       (sound_ay_registers[reg|1]<<8));
    if(!ay_tone_period[r])
      ay_tone_period[r]++;

    if(ay_tone_tick[r]>=ay_tone_period[r]*2)
      ay_tone_tick[r]%=ay_tone_period[r]*2;
    }
  
  /* generate tone */
  chan1=chan2=0;
  
  ay_tone_subcycles+=ay_tick_incr;
  tone_count=ay_tone_subcycles>>16;
  ay_tone_subcycles&=(1<<16)-1;
  
  if((sound_ay_registers[1]&128)==0)
    {
    level=AMPL_AY_TONE;
    AY_DO_TONE(chan1,0);
    }
  
  if((sound_ay_registers[3]&128)==0)
    {
    level=AMPL_AY_TONE;
    AY_DO_TONE(chan2,1);
    }

  /* write the sample(s) */
  (*ptr++)+=chan1+chan2;
  if(sound_stereo)
    (*ptr++)+=chan1+chan2;
  }
}


/* don't make the change immediately; record it for later,
 * to be made by sound_frame() (via sound_ay_overlay()).
 */
void sound_port_write(int port,int val,unsigned long tstates)
{
if(!sound_enabled) return;

if(port<0x50 || port>0x53)
  return;

if(ay_change_count<AY_CHANGE_MAX)
  {
  ay_change[ay_change_count].tstates=tstates;
  ay_change[ay_change_count].reg=port-0x50;
  ay_change[ay_change_count].val=val;
  ay_change_count++;
  }
}


/* no need to call this initially, but should be called
 * on reset otherwise.
 */
void sound_port_reset(void)
{
int f;

ay_change_count=0;
for(f=0;f<4;f++)
  sound_port_write(0x50+f,0,0);
for(f=0;f<2;f++)
  ay_tone_high[f]=0;
ay_tone_subcycles=0;
}


void sound_frame(void)
{
int fulllen=sound_framesiz*(sound_stereo+1);

/* XXX would be better to generate from scratch rather than overlay */
memset(sound_buf,128,fulllen);
sound_ay_overlay();

osssound_frame(sound_buf,fulllen);

ay_change_count=0;
}

#endif	/* OSS_SOUND_SUPPORT */
